# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from abc import ABCMeta, abstractmethod as abstract_method
from unittest import TestCase

from six import with_metaclass

from class_registry import AutoRegister, ClassRegistry


class AutoRegisterTestCase(TestCase):
    def test_auto_register(self):
        """
        Using :py:func:`AutoRegister` to, well, auto-register classes.
        """
        registry = ClassRegistry(attr_name='element')

        # Note that we declare :py:func:`AutoRegister` as the metaclass
        # for our base class.
        class BasePokemon(with_metaclass(AutoRegister(registry))):
            """
            Abstract base class; will not get registered.
            """
            @abstract_method
            def get_abilities(self):
                raise NotImplementedError()

        # noinspection PyClassHasNoInit
        class Sandslash(BasePokemon):
            """
            Non-abstract subclass; will get registered automatically.
            """
            element = 'ground'

            def get_abilities(self):
                return ['sand veil']

        # noinspection PyClassHasNoInit
        class BaseEvolvingPokemon(with_metaclass(ABCMeta, BasePokemon)):
            """
            Abstract subclass; will not get registered.
            """
            @abstract_method
            def evolve(self):
                raise NotImplementedError()

        # noinspection PyClassHasNoInit
        class Ekans(BaseEvolvingPokemon):
            """
            Non-abstract subclass; will get registered automatically.
            """
            element = 'poison'

            def get_abilities(self):
                return ['intimidate', 'shed skin']

            def evolve(self):
                return 'Congratulations! Your EKANS evolved into ARBOK!'


        self.assertListEqual(
            list(registry.items()),

            [
                # Note that only non-abstract classes got registered.
                ('ground', Sandslash),
                ('poison', Ekans),
            ],
        )

    def test_abstract_strict_definition(self):
        """
        If a class has no unimplemented abstract methods, it gets
        registered.
        """
        registry = ClassRegistry(attr_name='element')

        class FightingPokemon(with_metaclass(AutoRegister(registry))):
            element = 'fighting'

        self.assertListEqual(
            list(registry.items()),

            [
                # :py:class:`FightingPokemon` does not define any
                # abstract methods, so it is not considered to be
                # abstract!
                ('fighting', FightingPokemon),
            ],
        )

    def test_error_attr_name_missing(self):
        """
        The registry doesn't have an ``attr_name``.
        """
        registry = ClassRegistry()

        with self.assertRaises(ValueError):
            AutoRegister(registry)



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from .registry import *
from .auto_register import *
from .cache import *
from .entry_points import *
from .patcher import *



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals


class Pokemon(object):
    """
    A basic class with some attributes that we can use to test out
    class registries.
    """
    element = None

    def __init__(self, name=None):
        super(Pokemon, self).__init__()

        self.name = name


# Define some classes that we can register.
class Charmander(Pokemon):  element = 'fire'
class Charmeleon(Pokemon):  element = 'fire'
class Squirtle(Pokemon):    element = 'water'
class Wartortle(Pokemon):   element = 'water'
class Bulbasaur(Pokemon):   element = 'grass'
class Ivysaur(Pokemon):     element = 'grass'
class Mew(Pokemon):         element = 'psychic'


class PokemonFactory(object):
    """
    A factory that can produce new pokémon on demand.  Used to test how
    registries behave when a function is registered instead of a class.
    """
    @classmethod
    def create_psychic_pokemon(cls, name=None):
        return Mew(name)



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from abc import ABCMeta, abstractmethod as abstract_method
from collections import OrderedDict
from functools import cmp_to_key
from inspect import isclass as is_class
from typing import Any, Callable, Generator, Hashable, Iterator, \
    Mapping, MutableMapping, Optional, Text, Tuple, Union

from six import PY2, iteritems, with_metaclass

__all__ = [
    'BaseRegistry',
    'ClassRegistry',
    'MutableRegistry',
    'RegistryKeyError',
    'SortedClassRegistry',
]


class RegistryKeyError(KeyError):
    """
    Used to differentiate a registry lookup from a standard KeyError.

    This is especially useful when a registry class expects to extract
    values from dicts to generate keys.
    """
    pass


class BaseRegistry(with_metaclass(ABCMeta, Mapping)):
    """
    Base functionality for registries.
    """
    def __dir__(self):
        return list(self.keys())

    def __getitem__(self, key):
        """
        Shortcut for calling :py:meth:`get` with empty args/kwargs.
        """
        return self.get(key)

    def __iter__(self):
        # type: () -> Generator[Hashable]
        """
        Returns a generator for iterating over registry keys, in the
        order that they were registered.
        """
        return self.keys()

    @abstract_method
    def __len__(self):
        # type: () -> int
        """
        Returns the number of registered classes.
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )

    def __missing__(self, key):
        """
        Returns the result (or raises an exception) when the requested
        key can't be found in the registry.
        """
        raise RegistryKeyError(key)

    @abstract_method
    def get_class(self, key):
        """
        Returns the class associated with the specified key.
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )

    def get(self, key, *args, **kwargs):
        """
        Creates a new instance of the class matching the specified key.

        :param key:
            The corresponding load key.

        :param args:
            Positional arguments passed to class initializer.
            Ignored if the class registry was initialized with a null
            template function.

        :param kwargs:
            Keyword arguments passed to class initializer.
            Ignored if the class registry was initialized with a null
            template function.

        References:
          - :py:meth:`__init__`
        """
        return self.create_instance(self.get_class(key), *args, **kwargs)

    @staticmethod
    def gen_lookup_key(key):
        # type: (Any) -> Hashable
        """
        Used by :py:meth:`get` to generate a lookup key.

        You may override this method in a subclass, for example if you
        need to support legacy aliases, etc.
        """
        return key

    @staticmethod
    def create_instance(class_, *args, **kwargs):
        # type: (type, ...) -> Any
        """
        Prepares the return value for :py:meth:`get`.

        You may override this method in a subclass, if you want to
        customize the way new instances are created.

        :param class_:
            The requested class.

        :param args:
            Positional keywords passed to :py:meth:`get`.

        :param kwargs:
            Keyword arguments passed to :py:meth:`get`.
        """
        return class_(*args, **kwargs)

    @abstract_method
    def items(self):
        # type: () -> Generator[Tuple[Hashable, type]]
        """
        Iterates over registered classes and their corresponding keys,
        in the order that they were registered.

        Note: For compatibility with Python 3, this method should
        return a generator.
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )

    def keys(self):
        # type: () -> Generator[Hashable]
        """
        Returns a generator for iterating over registry keys, in the
        order that they were registered.

        Note: For compatibility with Python 3, this method should
        return a generator.
        """
        for item in self.items():
            yield item[0]

    def values(self):
        # type: () -> Generator[type]
        """
        Returns a generator for iterating over registered classes, in
        the order that they were registered.

        Note: For compatibility with Python 3, this method should
        return a generator.
        """
        for item in self.items():
            yield item[1]

    # Add some compatibility aliases to make class registries behave
    # more like dicts in Python 2.
    if PY2:
        def iteritems(self):
            """
            Included for compatibility with :py:data:`six.iteritems`.
            Do not invoke directly!
            """
            return self.items()

        def iterkeys(self):
            """
            Included for compatibility with :py:data:`six.iterkeys`.
            Do not invoke directly!
            """
            return self.keys()

        def itervalues(self):
            """
            Included for compatibility with :py:data:`six.itervalues`.
            Do not invoke directly!
            """
            return self.values()


class MutableRegistry(with_metaclass(ABCMeta, BaseRegistry, MutableMapping)):
    """
    Extends :py:class:`BaseRegistry` with methods that can be used to
    modify the registered classes.
    """
    def __init__(self, attr_name=None):
        # type: (Optional[Text]) -> None
        """
        :param attr_name:
            If provided, :py:meth:`register` will automatically detect
            the key to use when registering new classes.
        """
        super(MutableRegistry, self).__init__()

        self.attr_name = attr_name

    def __delitem__(self, key):
        # type: (Hashable) -> None
        """
        Provides alternate syntax for unregistering a class.
        """
        self._unregister(key)

    def __repr__(self):
        return '{type}({attr_name!r})'.format(
            attr_name   = self.attr_name,
            type        = type(self).__name__,
        )

    def __setitem__(self, key, class_):
        # type: (Text, type) -> None
        """
        Provides alternate syntax for registering a class.
        """
        self._register(key, class_)

    def register(self, key):
        """
        Decorator that registers a class with the registry.

        Example::

           registry = ClassRegistry(attr_name='widget_type')

           @registry.register
           class CustomWidget(BaseWidget):
             widget_type = 'custom'
             ...

           # Override the registry key:
           @registry.register('premium')
           class AdvancedWidget(BaseWidget):
             ...

        :param key:
            The registry key to use for the registered class.
            Optional if the registry's :py:attr:`attr_name` is set.
        """
        if is_class(key):
            if self.attr_name:
                # Note that ``getattr`` will raise an AttributeError if
                # the class doesn't have the required attribute.
                self._register(getattr(key, self.attr_name), key)
                return key
            else:
                raise ValueError('Registry key is required.')

        def _decorator(cls):
            self._register(key, cls)
            return cls
        return _decorator

    def unregister(self, key):
        # type: (Any) -> type
        """
        Unregisters the class with the specified key.

        :param key:
            The registry key to remove (not the registered class!).

        :return:
            The class that was unregistered.

        :raise:
            - :py:class:`KeyError` if the key is not registered.
        """
        return self._unregister(self.gen_lookup_key(key))

    @abstract_method
    def _register(self, key, class_):
        # type: (Hashable, type) -> None
        """
        Registers a class with the registry.
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )

    @abstract_method
    def _unregister(self, key):
        # type: (Hashable) -> type
        """
        Unregisters the class at the specified key.
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )



class ClassRegistry(MutableRegistry):
    """
    Maintains a registry of classes and provides a generic factory for
    instantiating them.
    """
    def __init__(self, attr_name=None, unique=False):
        # type: (Optional[Text], bool) -> None
        """
        :param attr_name:
            If provided, :py:meth:`register` will automatically detect
            the key to use when registering new classes.

        :param unique:
            Determines what happens when two classes are registered with
            the same key:

            - ``True``: The second class will replace the first one.
            - ``False``: A ``ValueError`` will be raised.
        """
        super(ClassRegistry, self).__init__(attr_name)

        self.unique = unique

        self._registry = OrderedDict()

    def __len__(self):
        # type: () -> int
        """
        Returns the number of registered classes.
        """
        return len(self._registry)

    def __repr__(self):
        return '{type}(attr_name={attr_name!r}, unique={unique!r})'.format(
            attr_name   = self.attr_name,
            type        = type(self).__name__,
            unique      = self.unique,
        )

    def get_class(self, key):
        """
        Returns the class associated with the specified key.
        """
        lookup_key = self.gen_lookup_key(key)

        try:
            return self._registry[lookup_key]
        except KeyError:
            return self.__missing__(lookup_key)

    def items(self):
        # type: () -> Iterator[Tuple[Hashable, type]]
        """
        Iterates over all registered classes, in the order they were
        added.
        """
        return iteritems(self._registry)

    def _register(self, key, class_):
        # type: (Hashable, type) -> None
        """
        Registers a class with the registry.
        """
        if key in ['', None]:
            raise ValueError(
                'Attempting to register class {cls} '
                'with empty registry key {key!r}.'.format(
                    cls = class_.__name__,
                    key = key,
                ),
            )

        if self.unique and (key in self._registry):
            raise RegistryKeyError(
                '{cls} with key {key!r} is already registered.'.format(
                    cls = class_.__name__,
                    key = key,
                ),
            )

        self._registry[key] = class_

    def _unregister(self, key):
        # type: (Hashable) -> type
        """
        Unregisters the class at the specified key.
        """
        return (
            self._registry.pop(key)
                if key in self._registry
                else self.__missing__(key)
        )



class SortedClassRegistry(ClassRegistry):
    """
    A ClassRegistry that uses a function to determine sort order when
    iterating.
    """
    def __init__(
            self,
            sort_key,                   # type: Union[Text, Callable[[Tuple[Hashable, type], Tuple[Hashable, type]], int]]
            attr_name       = None,     # type: Optional[Text]
            unique          = False,    # type: bool
            reverse         = False,    # type: bool
    ):
        """
        :param sort_key:
            Attribute name or callable, used to determine the sort value.

            If callable, must accept two tuples of (key, class).

        :param attr_name:
            If provided, :py:meth:`register` will automatically detect
            the key to use when registering new classes.

        :param unique:
            Determines what happens when two classes are registered with
            the same key:

            - ``True``: The second class will replace the first one.
            - ``False``: A ``ValueError`` will be raised.

        :param reverse:
            Whether to reverse the sort ordering.
        """
        super(SortedClassRegistry, self).__init__(attr_name, unique)

        self._sort_key = (
            sort_key
                if callable(sort_key)
                else self.create_sorter(sort_key)
        )

        self.reverse = reverse

    def items(self):
        # type: () -> Iterator[Tuple[Hashable, type]]
        return sorted(
            iteritems(self._registry),
                key     = self._sort_key,
                reverse = self.reverse,
        )

    @staticmethod
    def create_sorter(sort_key):
        """
        Given a sort key, creates a function that can be used to sort
        items when iterating over the registry.
        """
        def sorter(a, b):
            # type: (Tuple[Hashable, type], Tuple[Hashable, type]) -> int
            a_attr = getattr(a[1], sort_key)
            b_attr = getattr(b[1], sort_key)

            return (a_attr > b_attr) - (a_attr < b_attr)

        return cmp_to_key(sorter)



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from typing import Dict, Optional, Text

from pkg_resources import iter_entry_points
from six import iteritems

from class_registry import BaseRegistry

__all__ = [
    'EntryPointClassRegistry',
]


class EntryPointClassRegistry(BaseRegistry):
    """
    A class registry that loads classes using setuptools entry points.
    """
    def __init__(self, group, attr_name=None):
        # type: (Text, Optional[Text]) -> None
        """
        :param group:
            The name of the entry point group that will be used to load
            new classes.

        :param attr_name:
            If set, the registry will "brand" each class with its
            corresponding registry key.  This makes it easier to
            perform reverse lookups later.

            Note: if a class already defines this attribute, the
            registry will overwrite it!
        """
        super(EntryPointClassRegistry, self).__init__()

        self.attr_name  = attr_name
        self.group      = group

        self._cache = None # type: Optional[Dict[Text, type]]
        """
        Caches registered classes locally, so that we don't have to
        keep iterating over entry points.
        """

    def __len__(self):
        return len(self._get_cache())

    def __repr__(self):
        return '{type}(group={group!r})'.format(
            group   = self.group,
            type    = type(self).__name__,
        )

    def get_class(self, key):
        try:
            cls = self._get_cache()[key]
        except KeyError:
            cls = self.__missing__(key)

        if self.attr_name:
            setattr(cls, self.attr_name, key)

        return cls

    def items(self):
        return iteritems(self._get_cache())

    def refresh(self):
        """
        Purges the local cache.
        The next access attempt will reload all entry points.

        This is useful if you load a distribution at runtime.
        Otherwise, it probably serves no useful purpose.
        """
        self._cache = None

    def _get_cache(self):
        # type: () -> Dict[Text, type]
        """
        Populates the cache (if necessary) and returns it.
        """
        if self._cache is None:
            self._cache = {
                e.name: e.load()
                    for e in iter_entry_points(self.group)
            }

        # noinspection PyTypeChecker
        return self._cache



# coding=utf-8
# Not importing unicode_literals because in Python 2 distutils, some
# values are expected to be byte strings.
from __future__ import absolute_import, division, print_function

from codecs import StreamReader, open
from os.path import dirname, join, realpath

from setuptools import setup

cwd = dirname(realpath(__file__))

##
# Load long description for PyPi.
with open(join(cwd, 'README.rst'), 'r', 'utf-8') as f: # type: StreamReader
    long_description = f.read()


##
# Off we go!
setup(
    name = 'class-registry',
    description = 'Factory+Registry pattern for Python classes.',
    url = 'https://class-registry.readthedocs.io/',

    version = '2.0.2',

    packages = ['class_registry'],

    long_description = long_description,

    install_requires = [
        'six',
        'typing; python_version < "3.0"',
    ],

    test_suite    = 'test',
    test_loader   = 'nose.loader:TestLoader',
    tests_require = ['nose'],

    license = 'MIT',

    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],

    keywords = 'registry pattern',

    author          = 'Phoenix Zerin',
    author_email    = 'phoenix.zerin@eflglobal.com',
)
